from skretrieval.core import OpticalGeometry
from skretrieval.core.radianceformat import RadianceGridded
import sasktran as sk
from copy import copy
import numpy as np
from pathlib import Path
import logging
import xarray as xr


class Osiris(object):
    def __init__(self, scan_no, load_psf=False, apply_rsas=True, load_invalid=False, max_tangent_altitude=100000,
                 backend='l1services', l1_folder=None, l1_file=None, apply_uv_shift=False):
        if backend.lower() == 'l1services':
            valid = not load_invalid
            self.spectrograph_data = self._load_from_l1services(scan_no, load_psf, apply_rsas, valid)
        elif backend.lower() == 'netcdf':
            self.spectrograph_data = self._load_from_netcdf(l1_folder, l1_file, scan_no, load_psf=load_psf, apply_uv_shift=apply_uv_shift)

        t_alt_diff = np.diff(self.spectrograph_data.tangent_altitude.values)

        self.scan_direction = np.nanmean(t_alt_diff > 0) > 0.5

        if self.scan_direction:
            self.scan_direction = 1
        else:
            self.scan_direction = -1

        if self.scan_direction == -1:
            self.spectrograph_data = self.spectrograph_data.sel(mjd=self.spectrograph_data.mjd.values[::-1])

        good_alts = self.spectrograph_data.tangent_altitude.values < max_tangent_altitude

        self.spectrograph_data = self.spectrograph_data.isel(mjd=good_alts)

        # Sometimes the scan edge detection can be off and we will not have ascending tangent altitudes, cut off these
        # extra measurements
        t_alt_diff = np.diff(self.spectrograph_data['tangent_altitude'].values) > 0
        first_idx = np.where(t_alt_diff)[0][0]
        last_idx = np.where(t_alt_diff)[0][-1] + 2

        self.spectrograph_data = self.spectrograph_data.isel(mjd=slice(first_idx, last_idx))

        # sometimes scans can still be messed up, having non-monotonic tangent altitudes
        # I assume these are just bad scans
        if not all(np.diff(self.spectrograph_data.tangent_altitude) > 0):
            raise ValueError('OSIRIS Scan {} tangent altitudes are not monotonic'.format(scan_no))

    def _load_from_netcdf(self, l1_folder, l1_file, scan_no, load_psf=False, apply_uv_shift=False):
        if l1_folder is None and l1_file is None:
            raise ValueError('l1_folder must be set to use backend netcdf')
            
        if l1_file is None:
            root_folder = Path(l1_folder)

            possible_files = list(root_folder.rglob('OSIRIS_L1_Spectrograph_{}*'.format(str(scan_no//1000).zfill(6))))

            if len(possible_files) == 0:
                raise ValueError('Could not find orbit file for orbit {}'.format(str(scan_no//1000)))
            if len(possible_files) > 1:
                raise ValueError('Found multiple possible files for orbit {}'.format(str(scan_no//1000)))

            l1_file = possible_files[0]

        spectrograph_data = xr.open_dataset(l1_file, group='scan_{}'.format(scan_no))

        mjd = np.array([t.to_julian_date() for t in spectrograph_data['time'].to_dataframe()['time']]) - 2400000.5

        spectrograph_data['mjd'] = (['time'], mjd)

        spectrograph_data = spectrograph_data.swap_dims({'time': 'mjd'})

        if load_psf:
            low_idx = spectrograph_data.wavelength.values < 470
            spectrograph_data.spectral_fwhm.values[low_idx] = float(spectrograph_data.uv_fwhm)
        
        if apply_uv_shift:
            low_idx = spectrograph_data.wavelength.values < 470
            spectrograph_data.spectral_fwhm.values[low_idx] = float(spectrograph_data.uv_fwhm)

            spectrograph_data.wavelength.values[low_idx] -= float(spectrograph_data.uv_wavelength_shift)

        spectrograph_data = spectrograph_data.rename({'spectral_fwhm': 'fwhm',
                                                      'look_vector': 'osiris_look_ecef',
                                                      'observer_position': 'osiris_position_ecef',
                                                      'radiance': 'data',
                                                      'radiance_noise': 'error'})

        return spectrograph_data


    def _load_from_l1services(self, scan_no, load_psf, apply_rsas, valid):
        import osirisl1services.readlevel1 as readlevel1
        import osirisl1services.services as services
        import osirisl1services.spatialrotation as sr

        spectrograph_data = readlevel1.open_level1_spectrograph(scan_no, load_psf=load_psf, apply_rsas=apply_rsas,
                                                                     valid=valid)

        spectrograph_data['osiris_look_ecef'] = spectrograph_data.l1.look_ecef
        spectrograph_data['osiris_position_ecef'] = spectrograph_data.l1.position_ecef
        spectrograph_data['tangent_altitude'] = (['mjd'], spectrograph_data.l1.altitude.values)

        # TODO: Adjust MJD based on exposure time?

        return spectrograph_data

    def sasktran_geometry(self):
        geometry = sk.Geometry()

        for mjd, sliced in self.spectrograph_data.groupby('mjd'):
            pos = sliced.osiris_position_ecef.values.flatten()
            look = sliced.osiris_look_ecef.values.flatten()

            geometry.lines_of_sight.append(sk.LineOfSight(observer=pos, look_vector=look, mjd=mjd))

        geometry.lines_of_sight = geometry.lines_of_sight[::self.scan_direction]

        return geometry

    def optical_axis(self):
        axis = []

        for mjd, sliced in self.spectrograph_data.groupby('mjd'):
            pos = sliced.osiris_position_ecef.values.flatten()
            look = sliced.osiris_look_ecef.values.flatten()

            local_up = pos / np.linalg.norm(pos)

            axis.append(OpticalGeometry(observer=np.array(pos), look_vector=np.array(look), local_up=local_up, mjd=mjd))

        axis = axis[::self.scan_direction]

        return axis

    def skretrieval_l1(self, remove_sun=False):
        new_data = self.spectrograph_data.copy(deep=True)

        mjd = new_data.mjd.values
        
        new_data = new_data.assign_coords({'mjd': np.arange(0, len(mjd))})

        new_data = new_data.rename({'osiris_look_ecef': 'los_vectors', 'osiris_position_ecef': 'observer_position',
                                    'data': 'radiance', 'mjd': 'los',
                                    'error': 'radiance_noise'})

        new_data['mjd'] = (['los'], mjd)

        if remove_sun:
            solar_spectrum = sk.SolarSpectrum()

            irradiance = solar_spectrum.irradiance(new_data.wavelength.values, mjd=np.nanmean(mjd), fwhm=new_data.fwhm.values)

            new_data['radiance'] /= irradiance
            new_data['radiance_noise'] /= irradiance

        new_data = new_data.swap_dims({'pixel': 'wavelength'})
        return RadianceGridded(new_data)

if __name__ == "__main__":
    os = Osiris(6432012)

    geo = os.sasktran_geometry()

    l1_data = os.skretrieval_l1(remove_sun=True)

    axis = os.optical_axis()

    print(axis)