from osiris_skretrieval.osiris import Osiris
import sasktran as sk
import numpy as np
from scipy import interpolate
import skretrieval.util as util
from skretrieval.core.lineshape import DeltaFunction, Rectangle, Gaussian
from skretrieval.core.sensor.spectrograph import SpectrographFast, _set_join
import skretrieval.core.radianceformat as radianceformat
import xarray as xr
from abc import ABC, abstractmethod


class OSIRISModelBase(ABC):
    def __init__(self, osiris: Osiris, atmosphere: sk.Atmosphere, options=None,
                 grid_spacing=1000, mask_osiris_bad_pixels=True):
        self._measurement_optical_axis = osiris.optical_axis()
        self._measurement_l1 = osiris.skretrieval_l1()
        self._fwhm = self._measurement_l1.data['fwhm']

        self._measurement_geometry = osiris.sasktran_geometry()

        self._model_geometry = self._calculate_model_geometry()
        self._atmosphere = atmosphere
        self._options = options
        self._grid_spacing = grid_spacing
        self._mask_osiris_bad_pixels = mask_osiris_bad_pixels

        self.engine = self._create_engine()
        self.engine.grid_spacing = grid_spacing

    @abstractmethod
    def _calculate_model_geometry(self) -> sk.Geometry:
        pass

    @abstractmethod
    def _create_engine(self) -> sk.Engine:
        pass

    @abstractmethod
    def _model_radiance(self) -> radianceformat.RadianceGridded:
        pass

    def calculate_radiance(self):
        radiance = self._model_radiance()

        if self._atmosphere.wf_species is not None:
            radiance.data = radiance.data.rename({'wf_{}'.format(self._atmosphere.wf_species): 'wf'})
        data = radiance

        data.data['fwhm'] = self._fwhm

        if self._mask_osiris_bad_pixels:
            data.data['radiance'] = data.data['radiance'].where(~np.isnan(self._measurement_l1.data['radiance']))

        return data


class OSIRISModel(object):
    def __init__(self, osiris: Osiris, atmosphere: sk.Atmosphere, model_vertical_resolution_rad=0.0002, options=None, model_wavel=None,
                 meas_wavel=None, grid_spacing=1000):
        # Calculate the scan rate in rad/s
        geodetic = sk.Geodetic()
        self.measurement_optical_axis = osiris.optical_axis()
        self.measurement_l1 = osiris.skretrieval_l1()
        tangent_locations = self.measurement_l1.tangent_locations()
        tangent_angles = []
        tangent_mjd = []
        for idx, o in enumerate(self.measurement_optical_axis):
            tangent = tangent_locations.isel(los=idx)

            geodetic.from_lat_lon_alt(float(tangent.latitude.values), float(tangent.longitude.values), float(tangent.altitude.values))

            tangent_radius = np.linalg.norm(geodetic.location)

            obs_radius = np.linalg.norm(o.observer)

            tangent_angles.append(np.arcsin(tangent_radius / obs_radius))
            tangent_mjd.append(o.mjd)

        scan_rate = np.polyfit(tangent_mjd, tangent_angles, deg=1)[0]

        self.scan_rate = np.abs(scan_rate / (24 * 3600)) # To rad/s

        # Set up high resolution vertical geometry
        self.measurement_geometry = osiris.sasktran_geometry()
        self.model_geometry = sk.Geometry()

        lower_los = self.measurement_geometry.lines_of_sight[0]
        upper_los = self.measurement_geometry.lines_of_sight[-1]

        rotation_axis = np.cross(lower_los.look_vector, upper_los.look_vector)
        rotation_axis /= np.linalg.norm(rotation_axis)

        # Need to get the measurement angles to interpolate the mjd's properly
        all_angles = np.array([np.arccos(np.dot(l.look_vector, lower_los.look_vector)) for l in self.measurement_geometry.lines_of_sight])
        all_angles[0] = 0
        all_mjd = np.array([l.mjd for l in self.measurement_geometry.lines_of_sight])
        all_observer = np.array([l.observer for l in self.measurement_geometry.lines_of_sight])

        mjd_interpolator = interpolate.interp1d(all_angles, all_mjd, bounds_error=False, fill_value='extrapolate')
        observer_interpolater = interpolate.interp1d(all_angles, all_observer, axis=0, bounds_error=False,
                                                     fill_value='extrapolate')

        fov_bounds = [0, 0]

        model_angles = np.arange(all_angles[0] + fov_bounds[0],
                                 all_angles[-1] + fov_bounds[1] + model_vertical_resolution_rad,
                                 model_vertical_resolution_rad)

        for model_angle in model_angles:
            rot_matrix = util.rotation_matrix(rotation_axis, model_angle)

            look_vector = rot_matrix @ lower_los.look_vector

            mjd = np.asscalar(mjd_interpolator(model_angle))
            obs = observer_interpolater(model_angle)

            self.model_geometry.lines_of_sight.append(sk.LineOfSight(mjd, obs, look_vector))

        self.spectrographs = []

        all_vert_fov = []
        for idx, o in enumerate(self.measurement_optical_axis):
            vert_fov = float(self.measurement_l1.data.isel(los=idx).exposureTime.values) * self.scan_rate
            os_pixel = DeltaFunction()

            sensor = SpectrographFast(meas_wavel, os_pixel,
                                      vert_fov=Rectangle(width=vert_fov, mode='linear'),
                                      horiz_fov=Rectangle(width=1, mode='constant'))
            # sensor = SpectrographFast(model_wavel, os_pixel,
            #                           vert_fov=Gaussian(fwhm=0.0015502680433856196, mode='linear'),
            #                           horiz_fov=Rectangle(width=1, mode='constant'))

            self.spectrographs.append(sensor)

            all_vert_fov.append(vert_fov)

        self.model_wavel_nm = np.array(model_wavel)

        self.engine = sk.EngineHR(geometry=self.model_geometry, atmosphere=atmosphere, options=options)

        self.engine.wavelengths = self.model_wavel_nm

        self.engine.grid_spacing = grid_spacing

        self._fwhm = osiris.skretrieval_l1().data['fwhm']

    def calculate_radiance(self):
        radiance = self.engine.calculate_radiance()
        if hasattr(radiance, 'weighting_function'):
            wf = radiance.weighting_function
            radiance = radiance.radiance
        else:
            wf = None

        model_values = [s.model_radiance(o, self.model_wavel_nm, self.model_geometry, radiance, wf) for o, s
                        in zip(self.measurement_optical_axis, self.spectrographs)]

        data = radianceformat.RadianceGridded(xr.concat([m.data for m in model_values], dim='los'))

        data.data['fwhm'] = self._fwhm

        return data


class IdealOSIRISModel(OSIRISModelBase):
    def __init__(self, osiris: Osiris, atmosphere: sk.Atmosphere, options=None, meas_wavel=None, mask_osiris_bad_pixels=True,
                 grid_spacing=1000):
        super().__init__(osiris=osiris, atmosphere=atmosphere, options=options, mask_osiris_bad_pixels=mask_osiris_bad_pixels,
                         grid_spacing=grid_spacing)

        self.engine.wavelengths = meas_wavel

    def _calculate_model_geometry(self) -> sk.Geometry:
        return self._measurement_geometry

    def _create_engine(self) -> sk.Engine:
        return sk.EngineHR(geometry=self._model_geometry, atmosphere=self._atmosphere, options=self._options)

    def _model_radiance(self) -> radianceformat.RadianceGridded:
        rad = self.engine.calculate_radiance(output_format='xarray')

        return radianceformat.RadianceGridded(rad.rename({'los_vector': 'los_vectors'}))


class OSIRISModelConvWavelength(OSIRISModelBase):
    def __init__(self, osiris: Osiris, atmosphere: sk.Atmosphere, options=None, meas_wavel=None,
                 ss_resolution=0.1, mask_osiris_bad_pixels=True,
                 grid_spacing=1000, max_conv_wavel=500,
                 max_model_spectral_fwhm=3,
                 use_ss_approx=True):
        self._max_fwhms = max_model_spectral_fwhm
        self._use_ss_approx = use_ss_approx
        self._fwhm = osiris.skretrieval_l1().data.fwhm
        self._max_conv_wavel = max_conv_wavel
        self._construct_wavelengths(meas_wavel, ss_resolution)

        super().__init__(osiris=osiris, atmosphere=atmosphere, options=options, mask_osiris_bad_pixels=mask_osiris_bad_pixels,
                         grid_spacing=grid_spacing)

        horiz_line = Rectangle(width=1)
        vert_line = DeltaFunction()
        pixel_lines = []
        for w in meas_wavel:
            pixel_lines.append(Gaussian(fwhm=float(self._fwhm.sel(wavelength=w, method='nearest'))))
        self._sensor = SpectrographFast(meas_wavel, pixel_lines, vert_line, horiz_line)

    def _construct_wavelengths(self, meas_wavel, ss_resolution):
        lb = []
        ub = []
        for w in meas_wavel[meas_wavel < self._max_conv_wavel]:
            fwhm = float(self._fwhm.sel(wavelength=w, method='nearest'))
            lb.append(w - self._max_fwhms*fwhm)
            ub.append(w + self._max_fwhms*fwhm)

        final_sets = _set_join(lb, ub)
        wavelengths = []
        for set in final_sets:
            wavelengths.append(np.arange(set[0], set[1], ss_resolution))
        wavelengths.append(meas_wavel[meas_wavel >= self._max_conv_wavel])
        wavelengths = np.concatenate(wavelengths)

        wavelengths = np.sort(np.unique(np.concatenate([wavelengths, meas_wavel])))

        self._ss_wavelengths = wavelengths

        # Construct the MS Wavelengths
        self._ms_wavelengths = []
        for w in meas_wavel[meas_wavel < self._max_conv_wavel]:
            fwhm = float(self._fwhm.sel(wavelength=w, method='nearest'))
            self._ms_wavelengths.append([w - fwhm/2, w, w + fwhm/2])
        self._ms_wavelengths.append(meas_wavel[meas_wavel >= self._max_conv_wavel])
        self._ms_wavelengths.append([np.nanmin(self._ss_wavelengths)])
        self._ms_wavelengths = np.sort(np.hstack(self._ms_wavelengths))

        self._ss_wavelengths = np.sort(np.unique(np.concatenate([self._ms_wavelengths, self._ss_wavelengths])))

        # self._ms_wavelengths = np.sort(np.concatenate([[np.nanmin(wavelengths)], meas_wavel, [353]]))
        # self._ms_wavelengths = np.sort(np.unique(np.concatenate([self._ss_wavelengths[::4], meas_wavel])))


    def _calculate_model_geometry(self) -> sk.Geometry:
        return self._measurement_geometry

    def _create_engine(self) -> sk.Engine:
        if self._use_ss_approx:
            return sk.EngineHRSSApprox(self._model_geometry, atmosphere=self._atmosphere,
                                       wavelengths=self._ss_wavelengths, options=self._options,
                                       ms_wavelengths=self._ms_wavelengths)
        else:
            return sk.EngineHR(self._model_geometry, atmosphere=self._atmosphere,
                               wavelengths=self._ss_wavelengths, options=self._options)

    def _model_radiance(self) -> radianceformat.RadianceGridded:
        radiance = self.engine.calculate_radiance(output_format='xarray')
        if 'wf_{}'.format(self._atmosphere.wf_species) in radiance:
            wf = radiance['wf_{}'.format(self._atmosphere.wf_species)].values
        radiance = radiance['radiance'].values

        model_values = [self._sensor.model_radiance(o, self._ss_wavelengths, self._model_geometry, radiance, wf) for o
                        in self._measurement_optical_axis]

        data = radianceformat.RadianceGridded(xr.concat([m.data for m in model_values], dim='los'))

        data.data = data.data.rename({'wf': 'wf_{}'.format(self._atmosphere.wf_species)})

        return data


if __name__ == "__main__":
    os = Osiris(41880012)

    os_model = OSIRISModel(os, sk.Atmosphere(), model_wavel=[350])